module hybrid.IconCache;

private {
	import hybrid.BoxPacker;
	import utils.Singleton;
	import hybrid.GuiRenderer;
	import maths.Vec;
	import utils.Memory;
	import hybrid.Texture;
}



class IconCache {
	mixin SingletonMix;
	
	
	/// get some texture space of the requested size. return the offsets and a Texture object.
	Texture get(vec2i size, out vec2i bl, out vec2i tr, out vec2 blCoords, out vec2 trCoords) {
		auto block = packer.getBlock(size);
		if (block.page >= textures.length) {
			assert (texMngr !is null);
			textures ~= texMngr.createTexture(
					vec2i(texSize.x,	texSize.y),
					vec4(0, 0, 0, 0)
			);
		}
		
		bl = block.origin;
		tr = block.origin + block.size;
		
		blCoords = vec2(cast(float)bl.x / texSize.x, cast(float)bl.y / texSize.y);
		trCoords = vec2(cast(float)tr.x / texSize.x, cast(float)tr.y / texSize.y);

		// we'll shift the coords a bit so bilinear filtering doesn't kill us.
		blCoords += texelSize*0.0078125;
		trCoords += texelSize*0.0078125;
		
		return textures[block.page];
	}
	
	
	final vec2 texelSize() {
		return vec2(1.f / texSize.x, 1.f / texSize.y);
	}
	
	
	void initialize() {
		packer = new BoxPacker;
		packer.pageSize = texSize;
	}
	
	
	void updateTexture(Texture tex, vec2i origin, vec2i size, ubyte[] data) {
		return texMngr.updateTexture(tex, origin, size, data);
	}
	
	
	
	TextureMngr	texMngr;
	
	BoxPacker		packer;
	Texture[]		textures;
	
	vec2i			texSize = {x: 512, y: 512};
}


alias SingletonGetter!(IconCache) iconCache;
