module hybrid.Utils;

private {
	import maths.Vec;
}



struct Rect {
	vec2i min	= {x: int.max,	y: int.max};
	vec2i max	= {x: int.min,	y: int.min};


	int width()		{ return max.x - min.x; }
	int height()	{ return max.y - min.y; }
	

	static Rect opCall(vec2i min, vec2i max) {
		Rect res;
		res.min = min;
		res.max = max;
		return res;
	}
	
	
	vec2i size() {
		return vec2i(width, height);
	}
	
	
	bool intersect(Rect rect) {
		return	(min.x < rect.max.x && max.x > rect.min.x) &&
					(min.y < rect.max.y && max.y > rect.min.y);
	}
	
	
	bool contains(vec2i p) {
		return p.x >= min.x && p.x < max.x && p.y >= min.y && p.y < max.y;
	}
	
	
	void shrink(int x, int y) {
		min.x += x;
		min.y += y;
		max.x -= x;
		max.y -= y;
	}	

	static Rect sum(Rect a, Rect b) {
		int min(int a, int b)		{ return (a < b ? a : b); }
		int max(int a, int b)	{ return (a > b ? a : b); }
		
		Rect res;
		res.min.x	= min(a.min.x, b.min.x);
		res.min.y	= min(a.min.y, b.min.y);
		res.max.x	= max(a.max.x, b.max.x);
		res.max.y	= max(a.max.y, b.max.y);
		
		return res;
	}
	
	
	static Rect shaft(Rect a, Rect b) {
		int min(int a, int b)		{ return (a < b ? a : b); }
		int max(int a, int b)	{ return (a > b ? a : b); }
		
		Rect res;
		res.min.x	= min(a.max.x, b.max.x);
		res.max.x	= max(a.min.x, b.min.x);
		res.min.y	= min(a.max.y, b.max.y);
		res.max.y	= max(a.min.y, b.min.y);
		
		if (res.min.x > res.max.x) {
			int tmp = res.min.x;
			res.min.x = res.max.x;
			res.max.x = tmp;
		}

		if (res.min.y > res.max.y) {
			int tmp = res.min.y;
			res.min.y = res.max.y;
			res.max.y = tmp;
		}
		
		return res;
	}


	static Rect intersection(Rect a, Rect b) {
		int min(int a, int b)	{ return (a < b ? a : b); }
		int max(int a, int b)	{ return (a > b ? a : b); }
		
		Rect res;
		res.min.x	= max(a.min.x, b.min.x);
		res.max.x	= min(a.max.x, b.max.x);
		res.min.y	= max(a.min.y, b.min.y);
		res.max.y	= min(a.max.y, b.max.y);
		return res;
	}
}



struct Range {
	int from;
	int to;
	
	
	static Range opCall(int from, int to) {
		Range r;
		r.from = from;
		r.to = to;
		return r;
	}
}



char[] simpleAccessor(char[] name, char[] fieldName) {
	return "typeof(this) " ~ name ~ "(typeof(" ~ fieldName ~ ") o_) { " ~ fieldName ~ " = o_; return this; }"\n ~
	"typeof(" ~ fieldName ~ ") " ~ name ~ "() { return " ~ fieldName ~ "; }";
}
