﻿import hybrid.Font;
import hybrid.GlGuiRenderer;

import maths.Vec;

import derelict.sdl.sdl;
import derelict.util.loader;
import derelict.util.exception;
import derelict.opengl.gl;

import std.string, std.conv, std.file, std.stdio, std.c.time : usleep;



class SDLBackend : GlGuiRenderer {
	this() {
		Derelict_SetMissingProcCallback(function bool(char[] a, char[] b) { return true; });
			DerelictSDL.load();
		Derelict_SetMissingProcCallback(null);

		// initialize SDL's VIDEO module
		SDL_Init(SDL_INIT_VIDEO);
	
		// enable double-buffering
		SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
	
		// create our OpenGL window
		SDL_SetVideoMode(xResolution, yResolution, bitsPerPixel, SDL_OPENGL);
		SDL_WM_SetCaption(toStringz("Font rendering demo"), null);
		
		super.initialize();
	}
	

	bool updateWindow() {
        // swap the buffers, making our backbuffer the visible one
        SDL_GL_SwapBuffers();

		SDL_Event event;

        // handle all SDL events that we might've received in this loop iteration
        while (SDL_PollEvent(&event)) {
            switch (event.type) {
                // user has clicked on the window's close button
                case SDL_QUIT:
                    return false;

                // by default, we do nothing => break from the switch
                default:
                    break;
            }
        }

        // clear the screen. by default it clears to black
        glClear(GL_COLOR_BUFFER_BIT);

		return true;
	}
	
	
	override void cleanup() {
		super.cleanup();

		// tell SDL to quit
		SDL_Quit();
	
		DerelictSDL.unload();
	}
}



void main(char[][] args) {
	char[] displayText = import(`chinese.txt`).replace("\t", "    ");
	char[][] lines = splitlines(displayText);
	
	char[]	fontName	= `fonts/wt021.ttf`;
	int		fontSize	= 24;
	
	FontAntialiasing	antialiasing	= FontAntialiasing.LCD_RGB;
	LCDFilter			lcdFilter			= LCDFilter.Standard;

	argLoop: foreach (arg; args[1..$]) {
		switch (arg) {
			case `mono`:	antialiasing	= FontAntialiasing.None;			break;
			case `rgb`:		antialiasing	= FontAntialiasing.LCD_RGB;	break;
			case `bgr`:		antialiasing	= FontAntialiasing.LCD_BGR;	break;
			case `gray`:		antialiasing	= FontAntialiasing.Grayscale;	break;
			
			case `crisp`:		lcdFilter			= LCDFilter.Crisp;			break;
			case `std`:		lcdFilter			= LCDFilter.Standard;	break;
			
			default:
				if (arg.length > 5 && `font=` == arg[0..5]) {
					fontName = strip(arg[5..$]);
					continue argLoop;
				}

				if (arg.length > 5 && `size=` == arg[0..5]) {
					fontSize = toInt(strip(arg[5..$]));
					continue argLoop;
				}
			
				if (arg.length > 5 && `text=` == arg[0..5]) {
					lines = (cast(char[])std.file.read(strip(arg[5..$]))).replace("\t", "    ").splitlines();
					continue argLoop;
				}

				writefln(
`Usage: test8 [font=font_path] [size=font_size] [mono|rgb|bgr|gray] [crisp|std] [text=text_file]

	font - path to the font file
	size - height of the font face to use
	
	mono  - no antialiasing
	rgb   - subpixel antialiasing for horizontal RGB LCD subpixel layouts
	bgr   - subpixel antialiasing for horizontal BGR LCD subpixel layouts
	gray  - grayscale antialiasing
	
	crisp - crisp filter for subpixel antialiasing
	std   - standard filter for subpixel antialiasing
	
	text - path to a text file, whose contents will be displayed, UTF-8`);
				return;
		}
	}
	
	char[][] nonemptyLines;
	foreach (l; lines) {
		if (l.length != 0) nonemptyLines ~= l;
	}
	lines = nonemptyLines;


	auto fn = Font(fontName, fontSize);

	int screenH = 600 / fn.lineSkip;
	char[][] screen = new char[][screenH*4];
	float[] intens = new float[screenH*4];
	intens[] = 0.f;	
	
	
	auto renderer = new SDLBackend;

	fn.antialiasing = antialiasing;
	fn.lcdFilter = lcdFilter;
	
	while (renderer.updateWindow()) {
		{
			static int cursor = 0;
			static int off = 0;
			
			// print the lines, make sure to render brigher text on top of darker
			{
				int i = (cursor+1) % screen.length;
				for (int cnt = 0; cnt < screen.length; ++cnt) {
					float inten = intens[i];
					
					// dont bother to render the almost-black stuff
					if (inten > 0.01) {
						vec4 color = i == cursor ? vec4(1, 1, 1, 1) : vec4(inten*0.1, inten*1.0, inten*0.2, 1);
						fn.print(vec2i(2, (i % screenH) * fn.lineSkip), color, screen[i]);
					}

					++i;
					// flush before re-drawing text on top of other text, so we don't over-brighten the overlapping areas
					if (0 == i % screenH && i != 0) {
						fn.flush();
					}
					i %= screen.length;
				}
			}
			
			// darken the lines a bit
			foreach (inout i; intens) {
				i *= 0.96;
			}

			++cursor;
			cursor %= screen.length;
			
			++off;
			off %= lines.length;
			
			screen[cursor] = lines[off];
			intens[cursor] = 1;
		}

		// a horizontal text scroller
		{
			static int scrollPos = 800;
			char[] text = `D pwns yo momma! greetings: dbot, team0xf, #flipcode, #d, #d.tango, yo momma :P No fonts were harmed during the making of this demo.`;
			
			Font fn2 = Font(`fonts/Vera.ttf`, 72);
			int len = fn2.width(text);
			
			// flush the Chinese stuff, because we'll be doing some heavy overdraw here.
			fn2.flush();
			
			// first draw the shadow
			fn2.print(vec2i(5 + scrollPos, 5 + (600 - fn2.height)/2), vec4(0.5, 0.5, 0.5, 0.25), text);
			
			// again, overdraw
			fn2.flush();
			
			// now draw the real text ...
			fn2.print(vec2i(scrollPos, (600 - fn2.height)/2), vec4(0.5, 0.5, 0.5, 0.75), text);
			
			// ... and scroll
			scrollPos -= 5;
			if (scrollPos < -len) scrollPos = 800;
		}
		
		// a simple short string with a shadow
		{
			auto fn2 = Font(`fonts/Vera.ttf`, 14);
			char[] about = `Font rendering demo by h3r3tic`;
			int len = fn2.width(about);
			fn2.print(vec2i(800-len-2, 4), vec4(0.2, 0.2, 0.2, 1), about);
			fn2.flush();
			fn2.print(vec2i(800-len-4, 2), vec4(1, 0, 0, 1), about);
		}


		// flush all rendering buffers
		renderer.drawAll();
		
		// we don't want to go too fast with the text
		usleep(15_000);
	}
}
